#include "Grammar.h"

#include <vector>
#include <map>
#include <set>
#include <iostream>


/* *************** */
/* * Constructor * */
/* *************** */
Grammar::Grammar(multimap<Token*, vector<Token*> > r, Token* ax) {
	axiom = ax;
	rules = r;
	
	// Building lexGram
	for (multimap<Token*, vector<Token*> >::iterator it = r.begin(); it != r.end(); ++it) {
		lexGram.insert((*it).first);
		
		for (vector<Token*>::iterator it2 = (*it).second.begin(); it2 != (*it).second.end(); ++it2)
			lexGram.insert(*it2);
	}
	
}

Grammar::~Grammar() {

}



/* ******************************** */
/* * Analyse prédictive ( LL(0) ) * */
/* ******************************** */

/*
 * Compute the attributes nullable, first and follow where:
 * - Nullable(u) = true iff u --*--> \epsilon
 * - First(u) = { a \in lexGram  | u --*--> v_1.a.v_2  /\  a is terminal /\ v_1 --*--> \epsilon }
 * - Follow(X) = { a \in lexGram | Z --*--> uXav}
 */
void Grammar::fst_follow() {
	
	// Sets that will be built through fixpoints
	
	// **** Initialisations
	bool stabilityReached = false;
	for (set<Token*>::iterator it = lexGram.begin(); it!=lexGram.end(); ++it)
		nullable[*it] = ((*it)->isEpsilon);
	for (set<Token*>::iterator it = lexGram.begin(); it!=lexGram.end(); ++it)
		if ((*it)->isTerminal && !(*it)->isEpsilon)
			first[*it].insert(*it);
	
	// **** Algorithm
	while(!stabilityReached) {
		stabilityReached = true;
		
		// Pour chaque production "X -> x_1 ... x_n" (X -> vecRule)
		for (multimap<Token*, vector<Token*> >::iterator it = rules.begin(); it!=rules.end(); ++it) {
			Token* X = it->first;
			vector<Token*> vecRule = it->second;
			
			// ** Gestion des "nullable" **
			
			// "Si x1 ... xn = \epsilon alors nullable[X] = true;"
			// "Si tous les xi sont nullable alors nullable[X] = true;"
			bool isAllNullable = true;
			for (vector<Token*>::iterator it2 = vecRule.begin(); it2!=vecRule.end(); ++it2)
				isAllNullable &= nullable[*it2];
			if (isAllNullable)
				if (!nullable[X]) {
					stabilityReached = false;
					nullable[X] = true;
				}
			
			// ** Gestion des "first" **
			
			// Si "x_1 ... x_{i-1}" sont nullable
			//		=> first[X] = first[X] U first[x_i]
			for (vector<Token*>::iterator it2 = vecRule.begin(); it2!=vecRule.end(); ++it2) {
				
				// On récupère l'ensemble first[x_i], puis on met à jour first[X]
				//   en surveillant s'il y a un élément rajouté qui n'était pas dans le set
				set<Token*> fstxi = first[*it2];
				
				for (set<Token*>::iterator itTemp = fstxi.begin(); itTemp!=fstxi.end(); ++itTemp) {
					pair<set<Token*>::iterator,bool> ret = first[X].insert(*itTemp);
					if (ret.second)	// Si l'élément n'existait pas avant, alors quelque chose a bougé
						stabilityReached = false;
				}
				
				if (!nullable[*it2])
					break;
			}
			
			// ** Gestion des "follow" **
			
			// Basic rule
			for (int i=1; i<vecRule.size(); i++) {
				set<Token*> fstVecRulei = first[vecRule[i]];
				for (set<Token*>::iterator it2 = fstVecRulei.begin(); it2 != fstVecRulei.end(); ++it2)
					follow[vecRule[i-1]].insert(*it2);
			}
			
			// Si "x_{i+1} ... x_n" sont nullable alors follow[x_i] = follow[x_i] U follow[X];
			for (vector<Token*>::reverse_iterator it2 = vecRule.rbegin(); it2!=vecRule.rend(); ++it2) {
				// (it2 pointe sur x_i)
				
				set<Token*> follX = follow[X];
				for (set<Token*>::iterator itTemp = follX.begin(); itTemp!=follX.end(); ++itTemp) {
					pair<set<Token*>::iterator,bool> ret = follow[*it2].insert(*itTemp);
					if (ret.second)	// Si l'élément n'existait pas avant, alors quelque chose a bougé
						stabilityReached = false;
				}
				
				if (!nullable[*it2])	// Si le x_i courant est le premier non-nullable
					break;
			}
			
			// Pour chaque i = 1, n
			//	Pour chaque j = i+1, n
			//	  Si "x_{i+1} ... x_{j−1}" sont nullable, alors follow[x_i] = follow[x_i] U first[x_j];
			// => On cherche donc tous les plus segments maximaux [x_{i+1};x_{j-1}}] nullable pour updater le x_i avec x_j.
			// 	  Si "i=0" ou "j=n", on update rien du tout (cas "j=n" traité juste avant)
			
			// Tout d'abord, on se positionne au premier x_i qui soit non nullable
			vector<Token*>::iterator fstNonNullable;
			bool atLeastOneNonNullable = false;
			for (vector<Token*>::iterator it2 = vecRule.begin(); it2!=vecRule.end(); ++it2) {
				if (!nullable[*it2]) {
					atLeastOneNonNullable = true;
					fstNonNullable = it2;
					break;
				}
			}
			
			// Puis, on fait la recherche à partir de fstNonNullable dans vecRule
			if (atLeastOneNonNullable) {
				bool wasNullable = false;	// true ssi le token avant
				Token* xi = *fstNonNullable;
				
				for (vector<Token*>::iterator it2 = fstNonNullable; it2!=vecRule.end(); ++it2) {
					if (wasNullable && !nullable[*it2]) {
						// Fin d'un segment: x_i = xi et x_j = *it2
						
						set<Token*> fstxj = first[*it2];
						for (set<Token*>::iterator itTemp = fstxj.begin(); itTemp!=fstxj.end(); ++itTemp) {
							pair<set<Token*>::iterator,bool> ret = follow[xi].insert(*itTemp);
							if (ret.second)	// Si l'élément n'existait pas avant, alors quelque chose a bougé
								stabilityReached = false;
						}
					}
					
					if (!nullable[*it2])
						xi = *it2;	// xi restera au début du segment
					
					wasNullable = nullable[*it2];
				}
			}
		}
		
		/* DEBUG
		cout << "**** End of while loop ****" << endl;
		cout << "* First (size: " << first.size() << ") = " << endl;
		cout << first << endl;
		cout << "* Follow (size: " << follow.size() << ") = " << endl;
		cout << follow << endl;
		cout << "* Nullable (size: " << nullable.size() << ") = " << endl;
		cout << nullable << endl;
		cout << endl;//*/
		
	}
	
	return;
}



// Compute and return the director of (A-> u) as a mapping from the rule to the set
map<pair<Token*, vector<Token*> >, set<Token*> > Grammar::director() {
	
	map<pair<Token*, vector<Token*> >, set<Token*> > res;	// All directors (being computed)
	
	for (multimap<Token*, vector<Token*> >::iterator it = rules.begin(); it!=rules.end(); ++it) {
		
		set<Token*> dirRule;	// Director(A->vecRule)
		Token* A = (*it).first;
		vector<Token*> vecRule = (*it).second;
		
		// ** Algorithme:
		// Director(A -> x_0 ... x_n) = [ U_{i | x_0...x_{i-1} nullable} First(x_i) ]  U [ Follow(A) | if x_0 ... x_n nullable ]
		bool allNullable = true;
		
		if (vecRule.size()!=0) {
			// On rajoute dans tous les cas First(x_0)
			set<Token*> fstx0 = first[vecRule[0]];
			for (set<Token*>::iterator it2 = fstx0.begin(); it2 != fstx0.end(); ++it2)
				dirRule.insert(*it2);
			
			if (!nullable[vecRule[0]])
				allNullable = false;
			else {
				// Si les x_0 ... x_i étaient nullable, rajouter First(x_i)
				for (int i=1; i<vecRule.size(); i++) {
					
					// Rajout des éléments de First(x_i)
					set<Token*> fstxi = first[vecRule[i]];
					for (set<Token*>::iterator it2 = fstxi.begin(); it2 != fstxi.end(); ++it2)
						dirRule.insert(*it2);
					
					// Si x_i non-nullble => allNullable=false + interrupt the for loop
					if (!nullable[vecRule[i]]) {
						allNullable = false;
						break;
					}
				}
			}
		}
		
		// Si tous les éléments de la règle étaient nullable, alors rajouter Follow(A)
		if (allNullable) {
			set<Token*> followA = follow[A];
			for (set<Token*>::iterator it2 = followA.begin(); it2 != followA.end(); ++it2)
				dirRule.insert(*it2);
		}
		
		
		// Update de "res" avec "dirRule"
		pair<Token*, vector<Token*> > ruleTemp = make_pair(A, vecRule);
		res[ruleTemp] = dirRule;
	}
	
	return res;
}


// Compute the table of predictive analysis T[X][a] = rule (type of rule = pair<Token*, vector<Token*> >)
// WARNING: the grammar must be LL(1)
map<Token*, map<Token*, pair<Token*, vector<Token*> > > > Grammar::tableAnalysePred(map<pair<Token*, vector<Token*> >, set<Token*> > direct) {
	
	map<Token*, map<Token*, pair<Token*, vector<Token*> > > > table;
	
	// ** Algorithme:
	// T[X][a] = (X -> u), pour chaque a \in Director(X → u)
	for (map<pair<Token*, vector<Token*> >, set<Token*> >::iterator it = direct.begin(); it!= direct.end(); ++it) {
		
		pair<Token*, vector<Token*> > rule = it->first;		// (X->u)
		Token* X = rule.first;
		set<Token*> dirRule = it->second;					// { a | a \in Dir(X->u) }
		
		for (set<Token*>::iterator it2 = dirRule.begin(); it2!=dirRule.end(); ++it2)
			(table[X])[*it2] = rule;
	}
	
	
	return table;
}



/* ****************** */
/* * Pretty printer * */
/* ****************** */
ostream& operator <<(ostream& sout, Grammar& gram) {
	sout << "Starting axiom : " << *(gram.axiom) << endl;
	sout << "Rules :" << endl;
	for (multimap<Token*, vector<Token*> >::iterator it= (gram.rules).begin(); it!=(gram.rules).end(); ++it) {
		sout << *(it->first) << " -> ";
		for (vector<Token*>::iterator it2 = (it->second).begin(); it2!=(it->second).end(); ++it2)
			sout << **it2 << " ";
		sout << endl;
	}
	sout << endl;
	
	return sout;
}


ostream& operator <<(ostream& sout, map<Token*, set<Token*> >& mapping) {
	for (map<Token*, set<Token*> >::iterator it = mapping.begin(); it != mapping.end(); ++it) {
		Token* arg = it->first;
		set<Token*> valueSet = it->second;
		
		sout << *arg << " -> {";
		for (set<Token*>::iterator it2 = valueSet.begin(); it2!=valueSet.end(); ++it2)
			sout << " " << *(*it2);
		sout << " }" << endl;
	}
	
	return sout;
}


ostream& operator <<(ostream& sout, map<Token*, bool >& mapping) {
	for (map<Token*, bool >::iterator it = mapping.begin(); it != mapping.end(); ++it) {
		Token* arg = it->first;
		bool value = it->second;
		
		sout << *arg << " -> ";
		if (value)
			sout << "true" << endl;
		else
			sout << "false" << endl;
	}
	
	return sout;
}

ostream& operator <<(ostream& sout, map<pair<Token*, vector<Token*> >, set<Token*> >& mapping) {
	for (map<pair<Token*, vector<Token*> >, set<Token*> >::iterator it = mapping.begin(); it != mapping.end(); ++it) {
		pair<Token*, vector<Token*> > rule = it->first;
		set<Token*> valueSet = it->second;
		
		sout << "( " << *(rule.first) << " ->";
		for (int i=0; i<rule.second.size(); i++)
			sout << " " << *((rule.second)[i]);
		sout << " ) \t\t=> {";
		for (set<Token*>::iterator it2 = valueSet.begin(); it2!=valueSet.end(); ++it2)
			sout << " " << *(*it2);
		sout << " }" << endl;
	}
	
	return sout;
}

ostream& operator <<(ostream& sout, map<Token*, map<Token*, pair<Token*, vector<Token*> > > >& mapping) {
	for (map<Token*, map<Token*, pair<Token*, vector<Token*> > > >::iterator it = mapping.begin(); it != mapping.end(); ++it) {
		Token* X = it->first;
		map<Token*, pair<Token*, vector<Token*> > > rest = it->second;
		
		for (map<Token*, pair<Token*, vector<Token*> > >::iterator it2 = rest.begin(); it2 != rest.end(); ++it2) {
			Token* a = it2->first;
			pair<Token*, vector<Token*> > rule = it2->second;
			
			sout << "Table" << *X << "[" << *a << "] = (" << *(rule.first) << " ->";
			for (int i=0; i<rule.second.size(); i++)
				sout << " " << *((rule.second)[i]);
			
			sout << " )" <<endl;
		}
	}
	sout << "Else, we have a parsing error."<< endl;
	
	return sout;
}

